"""The evaluator package"""

from texasholdem.evaluator.evaluator import (evaluate,
                                             get_rank_class,
                                             rank_to_string,
                                             get_five_card_rank_percentage)
